import sys
import json, ssl
import http.client
import urllib.parse
import random
import datetime
from time import gmtime, strftime, sleep
from arcgis.gis import GIS
from arcgis import features

# connect the Hue bridge to the router, then
# run this, then go and click the Hue bridge button and run it again - it should give you a username

# to name or rename the sensors, either do that to the source FGDB, or simply edit the current features' names in the AGOL current layer - those edits will be applied to the ongoing sensors

serverName = "192.168.26.129" # Getech London bridge - in London, added as a reservation to DHCP
serverPort = 80
serverPortSSL = 443
apiURL = "/api"
jsonCreateUser = {'devicetype':'dell3520'}
username = '<insertYourAPIUsername>'

# AGOL credentials to use when querying target service
agolUsername = "<insertYourAGOLUserName>"
agolPassword = "<insertYourAGOLPassword>"
# target AGOL server
agolServer = 'www.arcgis.com'
# target owner of items
targetOwner = "<insertYourAGOLOwnerUserName>"
targetTitle = "HueSensorDataLondonCurrentWGS"
targetTitleOngoing = "HueSensorDataLondonOngoingWGS"
targetTitleZones = "HueSensorZonesLondonCurrentWGS"
# refresh AGOL connection time in seconds
agolRefreshTime = 3600

sleepSeconds = 3
keepForDays = 30
#secondsPerDay = 86400
#deleteRecordsCount = int((secondsPerDay / sleepSeconds) * keepForDays)
deleteRecordsCount = 3600 / sleepSeconds
areTesting = False

# A function that checks that the input JSON object
#  is not an error object.
def assertJsonSuccess(data):
    obj = json.loads(data)
    if 'status' in obj and obj['status'] == "error":
        print("ERROR: JSON object returns an error: {0}".format((obj)))
        return False
    else:
        return True

def sendHTTPRequest(serverName, serverPort, requestURL, requestType, content):
    headers = {"Content-type": "application/x-www-form-urlencoded", "Accept": "text/plain", "allow_redirects":"True"}
      # this context allows SSL certificate warnings to be ignored
    #gContext = ssl._create_unverified_context()
    httpConn = http.client.HTTPConnection(serverName, serverPort)

    if requestType.upper() == "POST":
        if content == None:
            httpConn.request("POST", requestURL, "", headers)
        else:
            headers = {"Content-type": "application/x-www-form-urlencoded", "Accept": "text/plain"}
            httpConn.request("POST", requestURL, json.dumps(content).encode(), headers)
            #print(params)
    elif requestType.upper() == "PUT":
        httpConn.request("PUT", requestURL, json.dumps(content).encode(), headers)
    else:
        httpConn.request("GET", requestURL, "", headers)
# Read response
    response = httpConn.getresponse()
    print(response.status)
    if (response.status == 302):
        print("ERROR: redirect")
        httpConn.close()
        return
    elif (response.status == 200):
        data = response.read()
        httpConn.close()

        #print("INFO: Response data: {0}".format(data))
        # Check that data returned is not an error object
        if not assertJsonSuccess(data):
            return
        # Generate the JSON and return it
        json_response = json.loads(data)
        return json_response
    else:
        print("ERROR: Response code: {0}".format(response.status))
        httpConn.close()
        return

def sendHTTPSRequest(serverName, serverPort, requestURL, requestType, content):
    headers = {"Content-type": "application/x-www-form-urlencoded", "Accept": "text/plain", "allow_redirects":"True"}
      # this context allows SSL certificate warnings to be ignored
    gContext = ssl._create_unverified_context()
    httpsConn = http.client.HTTPSConnection(serverName, serverPort, context=gContext)

    if requestType.upper() == "POST":
        if content == None:
            httpsConn.request("POST", requestURL, "", headers)
        else:
            headers = {"Content-type": "application/x-www-form-urlencoded", "Accept": "text/plain"}
            httpsConn.request("POST", requestURL, json.dumps(content).encode(), headers)
    elif requestType.upper() == "PUT":
        httpsConn.request("PUT", requestURL, json.dumps(content).encode(), headers)
    else:
        httpsConn.request("GET", requestURL, "", headers)

    response = httpsConn.getresponse()
    if (response.status == 302):
        print("ERROR: redirect")
        httpsConn.close()
        return
    elif (response.status == 200):
        data = response.read()
        httpsConn.close()
        #print("INFO: Response data: {0}".format(data))
        # Check that data returned is not an error object
        if not assertJsonSuccess(data):
            return
        # Generate the JSON and return it
        json_response = json.loads(data)
        return json_response
    else:
        print("ERROR: Response code: {0}".format(response.status))
        httpsConn.close()
        return


class HueSensor:
    def __init__(self, id = None,  name = None, type = None, modelid = None, uniqueid = None, manufacturername = None, productname = None, daylight = None, presence = False, reachable = False, dark = False, lightlevel = -999, temperature = -999):
        self.id = id
        self.name = name
        self.type = type
        self.modelid = modelid
        self.uniqueid = uniqueid
        self.manufacturername = manufacturername
        self.productname = productname
        # type: Not Daylight
        self.reachable = reachable
        # type: Daylight, ZLLLightLevel
        self.daylight = daylight
        # type: ZLLLightLevel
        self.dark = dark
        self.lightlevel = lightlevel
        # type: ZLLPresence
        self.presence = presence
        # type: ZLLTemperature
        self.temperature = temperature


def main():
    #https://wolfgang-ziegler.com/blog/hue-motion-sensor-api
    sensors = sendHTTPRequest(serverName, serverPort, apiURL + "/" + username + "/sensors", "GET", "")

    getDataFromSensors = []

    for item in sensors:
        sensor = sensors[item]

        if 'type' in sensor:
            sensortype = sensor['type']
            #print(sensortype)
            if sensortype.lower() == 'zlllightlevel':
                getDataFromSensors.append(HueSensor(item, sensor['name'], sensor['type'], sensor['modelid'], sensor['uniqueid'], sensor['manufacturername'], sensor['productname'], sensor['state']['daylight'], None, sensor['config']['reachable'], sensor['state']['dark'], sensor['state']['lightlevel'], None))
                print("sensor name: {0}, id: {1}, modelid: {2}, productname: {3}, daylight: {4}, lightlevel: {5}, isdark: {6}, uniqueID: {7}".format(sensor['name'], item, sensor['modelid'], sensor['productname'], sensor['state']['daylight'], sensor['state']['lightlevel'], sensor['state']['dark'], sensor['uniqueid']))
            elif sensortype.lower() == 'zlltemperature':
                getDataFromSensors.append(HueSensor(item, sensor['name'], sensor['type'], sensor['modelid'], sensor['uniqueid'], sensor['manufacturername'], sensor['productname'], None, None, sensor['config']['reachable'], None, None, sensor['state']['temperature']))
                print("sensor name: {0}, id: {1}, modelid: {2}, productname: {3}, temp: {4}, uniqueID: {5}".format(sensor['name'], item, sensor['modelid'], sensor['productname'], sensor['state']['temperature'],sensor['uniqueid']))
            elif sensortype.lower() == 'zllpresence':
                getDataFromSensors.append(HueSensor(item, sensor['name'], sensor['type'], sensor['modelid'], sensor['uniqueid'], sensor['manufacturername'], sensor['productname'], None, sensor['state']['presence'], sensor['config']['reachable'], None, None, None))
                print("sensor name: {0}, id: {1}, modelid: {2}, productname: {3}, presence: {4}, uniqueID: {5}".format(sensor['name'], item, sensor['modelid'], sensor['productname'], sensor['state']['presence'],sensor['uniqueid']))

    doneItThisManyTimes = 0
    doItThisManyTimes = 1000

    startTime = datetime.datetime.utcnow()
    ct = 1
    agol = None
    # do this infinitely
    while ct != 0:
        timeNow = datetime.datetime.utcnow()
        if agol == None or (timeNow - startTime).seconds > agolRefreshTime:
            startTime = datetime.datetime.utcnow()
            print("refreshing AGOL connection: {0} (UTC)".format(timeNow.strftime('%m/%d/%y %H:%M:%S')))
            # connect to AGOL
            try:
                agol = GIS("https://" + agolServer, agolUsername, agolPassword)
            except:
                error_message =  str(sys.exc_info()[1]).split("\n")[0]
                print("ERROR: error when connecting to ArcGIS Online via proxy: {0}".format(error_message))
                sys.exit()

            # search for the target feature service for the 'current' point values - this is the layer you look at to see what's happening right now
            qry = 'title:"{0}" AND owner:"{1}" '.format(targetTitle, targetOwner)
            items = agol.content.search(query=qry)
            fsTarget = "Feature Service"
            presenceSensorLayer = None
            lightSensorLayer = None
            tempSensorLayer = None

            # identify the target layers within the target feature service
            for item in items:
                #print("INFO: item title: {0}, item type: {1}, item owner: {2}".format(item.title, item.type, item.owner))
                if item.title == targetTitle:
                    if item.type == fsTarget:
                        layers = item.layers
                        for layer in layers:
                            if layer.properties["name"] == "HuePresenceSensors":
                                presenceSensorLayer = layer
                            if layer.properties["name"] == "HueLightSensors":
                                lightSensorLayer = layer
                            if layer.properties["name"] == "HueTempSensors":
                                tempSensorLayer = layer

            # search for the target feature service for the 'ongoing' point values - these are the historic values we're keeping to allow us to rewind time and see what the temperature was, if there was someone there, etc
            qry = 'title:"{0}" AND owner:"{1}" '.format(targetTitleOngoing, targetOwner)
            items = agol.content.search(query=qry)
            fsTarget = "Feature Service"
            presenceSensorLayerOngoing = None
            lightSensorLayerOngoing = None
            tempSensorLayerOngoing = None

            for item in items:
                #print("INFO: item title: {0}, item type: {1}, item owner: {2}".format(item.title, item.type, item.owner))
                if item.title == targetTitleOngoing:
                    if item.type == fsTarget:
                        layers = item.layers
                        for layer in layers:
                            if layer.properties["name"] == "HuePresenceSensors":
                                presenceSensorLayerOngoing = layer
                            if layer.properties["name"] == "HueLightSensors":
                                lightSensorLayerOngoing = layer
                            if layer.properties["name"] == "HueTempSensors":
                                tempSensorLayerOngoing = layer


            # search for the target feature service for the 'current' zone values - this is the polygon layer which shows presence/absence in each zone
            qry = 'title:"{0}" AND owner:"{1}" '.format(targetTitleZones, targetOwner)
            items = agol.content.search(query=qry)
            fsTarget = "Feature Service"
            presenceSensorZoneLayer = None

            for item in items:
                if item.title == targetTitleZones:
                    if item.type == fsTarget:
                        layers = item.layers
                        for layer in layers:
                            if layer.properties["name"] == "HueSensorPresenceZones":
                                presenceSensorZoneLayer = layer

        try:
            # loop through the features in the current point feature layer - there's one feature for each sensor
            # do light level sensors
            lightSensorLayer_feats = lightSensorLayer.query().features
            edit_feats = lightSensorLayer_feats
            now = datetime.datetime.utcnow()
            fmtDT = now.strftime('%m/%d/%y %H:%M:%S')

            addRecords = []
            for sensor in getDataFromSensors:
                if sensor.type.lower() == 'zlllightlevel':
                    # query the Hue bridge to get the last light level value reported by each sensor
                    requestUrl = apiURL + "/" + username + "/sensors/" + sensor.id
                    resp = sendHTTPSRequest(serverName, serverPortSSL, requestUrl, "GET", "")
                    geom = None
                    name = None
                    sensorNumber = None
                    for f in edit_feats:
                         if f.attributes['UniqueID'] == sensor.uniqueid:
                            geom = f.geometry
                            name = f.attributes['Name']
                            sensorNumber = f.attributes['SensorNumber']
                            f.attributes['Daylight'] = resp['state']['daylight']
                            f.attributes['Dark'] = resp['state']['dark']
                            valToSet = resp['state']['lightlevel']
                            if resp['config']['reachable'] != True:
                                valToSet = -1
                            f.attributes['Lightlevel'] = valToSet
                            f.attributes['LastReading'] = fmtDT
                    # update the AGOL feature to show the current values
                    lightSensorLayer.edit_features(updates=edit_feats)

                    valToSet = resp['state']['lightlevel']
                    if resp['config']['reachable'] != True:
                        valToSet = -1

                    # add the new data to the ongoing data array and then update that layer too
                    addRecords.append({'attributes':{'UniqueID': sensor.uniqueid, 'Name': name, 'SensorNumber' : sensorNumber, 'Daylight': resp['state']['daylight'], 'Dark': resp['state']['dark'],'Lightlevel' : valToSet, 'ReadingDateTime':fmtDT}, 'geometry': {'x': geom['x'], 'y': geom['y']}})
            if len(addRecords) > 0:
                lightSensorLayerOngoing.edit_features(adds=addRecords)
        except:
            print("issue in lightlevel")

        try:
            # do presence sensors
            presenceSensorLayer_feats = presenceSensorLayer.query().features
            edit_feats = presenceSensorLayer_feats
            presenceSensorZoneLayer_feats = presenceSensorZoneLayer.query(return_geometry=False).features
            edit_zone_feats = presenceSensorZoneLayer_feats
            now = datetime.datetime.utcnow()
            fmtDT = now.strftime('%m/%d/%y %H:%M:%S')

            addRecords = []
            for sensor in getDataFromSensors:
                if sensor.type.lower() == 'zllpresence':
                    requestUrl = apiURL + "/" + username + "/sensors/" + sensor.id
                    resp = sendHTTPSRequest(serverName, serverPortSSL, requestUrl, "GET", "")
                    geom = None
                    name = None
                    sensorNumber = None
                    for f in edit_feats:
                         if f.attributes['UniqueID'] == sensor.uniqueid:
                            geom = f.geometry
                            name = f.attributes['Name']
                            sensorNumber = f.attributes['SensorNumber']
                            valToSet = resp['state']['presence']
                            if resp['config']['reachable'] != True:
                                valToSet = -1
                            f.attributes['Presence'] = valToSet
                            f.attributes['LastReading'] = fmtDT
                    presenceSensorLayer.edit_features(updates=edit_feats)
                    valToSet = resp['state']['presence']
                    if resp['config']['reachable'] != True:
                        valToSet = -1
                    addRecords.append({'attributes':{'UniqueID': sensor.uniqueid, 'Name': name, 'SensorNumber' : sensorNumber, 'Presence': valToSet, 'ReadingDateTime':fmtDT}, 'geometry': {'x': geom['x'], 'y': geom['y']}})
                    for f in edit_zone_feats:
                         if f.attributes['UniqueID'] == sensor.uniqueid:
                            valToSet = resp['state']['presence']
                            if resp['config']['reachable'] != True:
                                valToSet = -1
                            f.attributes['Presence'] = valToSet
                            f.attributes['LastReading'] = fmtDT
                    presenceSensorZoneLayer.edit_features(updates=edit_zone_feats)

            if len(addRecords) > 0:
                presenceSensorLayerOngoing.edit_features(adds=addRecords)

        except:
            print("issue in presence")

        try:
            # do temperature sensors
            tempSensorLayer_feats = tempSensorLayer.query().features
            edit_feats = tempSensorLayer_feats
            now = datetime.datetime.utcnow()
            fmtDT = now.strftime('%m/%d/%y %H:%M:%S')

            addRecords = []
            for sensor in getDataFromSensors:
                if sensor.type.lower() == 'zlltemperature':
                    requestUrl = apiURL + "/" + username + "/sensors/" + sensor.id
                    resp = sendHTTPSRequest(serverName, serverPortSSL, requestUrl, "GET", "")
                    geom = None
                    name = None
                    sensorNumber = None
                    for f in edit_feats:
                         if f.attributes['UniqueID'] == sensor.uniqueid:
                            geom = f.geometry
                            name = f.attributes['Name']
                            sensorNumber = f.attributes['SensorNumber']
                            valToSet = resp['state']['temperature']
                            if resp['config']['reachable'] != True:
                                valToSet = -100
                            valToSetCelsius =  valToSet / 100
                            scaledValue = valToSetCelsius
                            for scaleTemp in scaleTemps:
                                if scaleTemp[0] == sensorNumber:
                                    scaledValue = valToSetCelsius
                            f.attributes['Temperature'] = scaledValue
                            f.attributes['LastReading'] = fmtDT
                    tempSensorLayer.edit_features(updates=edit_feats)

                    valToSet = resp['state']['temperature']
                    valToSetCelsius = valToSet / 100
                    scaledValue = valToSetCelsius
                    for scaleTemp in scaleTemps:
                        if scaleTemp[0] == sensorNumber:
                            scaledValue = valToSetCelsius
                    if resp['config']['reachable'] != True:
                        scaledValue = -100
                    addRecords.append({'attributes':{'UniqueID': sensor.uniqueid, 'Name' : name, 'SensorNumber' : sensorNumber, 'Temperature': scaledValue, 'ReadingDateTime':fmtDT}, 'geometry': {'x': geom['x'], 'y': geom['y']}})
            if len(addRecords) > 0:
                tempSensorLayerOngoing.edit_features(adds=addRecords)
        except:
            print("issue in temperature")


        if doneItThisManyTimes % deleteRecordsCount == 0 or areTesting:
            try:
                print("INFO: attempting delete of old presence observations")
                presenceSensorLayerOngoing.delete_features(where="ReadingDateTime < CURRENT_TIMESTAMP - {0}".format(keepForDays))
                print("INFO: attempting delete of old light level observations")
                lightSensorLayerOngoing.delete_features(where="ReadingDateTime < CURRENT_TIMESTAMP - {0}".format(keepForDays))
                print("INFO: attempting delete of old temperature observations")
                tempSensorLayerOngoing.delete_features(where="ReadingDateTime < CURRENT_TIMESTAMP - {0}".format(keepForDays))
            except:
                error_message = str(sys.exc_info()[1]).split("\n")[0]
                print("ERROR: message: {0}".format(error_message))


        doneItThisManyTimes += 1
        sleep(sleepSeconds)

    return



if __name__ == '__main__':
    main()
